package com.github.k1rakishou.kurobaexlite.features.bookmarks

import android.content.Context
import androidx.activity.ComponentActivity
import androidx.compose.animation.Animatable
import androidx.compose.animation.core.FiniteAnimationSpec
import androidx.compose.animation.core.animateInt
import androidx.compose.animation.core.snap
import androidx.compose.animation.core.tween
import androidx.compose.animation.core.updateTransition
import androidx.compose.foundation.ExperimentalFoundationApi
import androidx.compose.foundation.Image
import androidx.compose.foundation.background
import androidx.compose.foundation.layout.BoxWithConstraints
import androidx.compose.foundation.layout.Column
import androidx.compose.foundation.layout.PaddingValues
import androidx.compose.foundation.layout.Row
import androidx.compose.foundation.layout.Spacer
import androidx.compose.foundation.layout.fillMaxSize
import androidx.compose.foundation.layout.fillMaxWidth
import androidx.compose.foundation.layout.height
import androidx.compose.foundation.layout.padding
import androidx.compose.foundation.layout.size
import androidx.compose.foundation.layout.width
import androidx.compose.foundation.lazy.LazyItemScope
import androidx.compose.foundation.lazy.rememberLazyListState
import androidx.compose.foundation.text.InlineTextContent
import androidx.compose.foundation.text.appendInlineContent
import androidx.compose.runtime.Composable
import androidx.compose.runtime.DisposableEffect
import androidx.compose.runtime.LaunchedEffect
import androidx.compose.runtime.State
import androidx.compose.runtime.collectAsState
import androidx.compose.runtime.derivedStateOf
import androidx.compose.runtime.getValue
import androidx.compose.runtime.mutableStateOf
import androidx.compose.runtime.remember
import androidx.compose.runtime.setValue
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.draw.drawBehind
import androidx.compose.ui.graphics.Color
import androidx.compose.ui.graphics.graphicsLayer
import androidx.compose.ui.layout.ContentScale
import androidx.compose.ui.platform.LocalContext
import androidx.compose.ui.platform.LocalDensity
import androidx.compose.ui.res.dimensionResource
import androidx.compose.ui.res.painterResource
import androidx.compose.ui.text.AnnotatedString
import androidx.compose.ui.text.Placeholder
import androidx.compose.ui.text.PlaceholderVerticalAlign
import androidx.compose.ui.text.SpanStyle
import androidx.compose.ui.text.buildAnnotatedString
import androidx.compose.ui.text.style.TextOverflow
import androidx.compose.ui.unit.dp
import androidx.compose.ui.unit.sp
import coil.compose.AsyncImagePainter
import coil.compose.SubcomposeAsyncImage
import coil.compose.SubcomposeAsyncImageContent
import coil.request.ImageRequest
import coil.size.Size
import coil.transform.CircleCropTransformation
import com.github.k1rakishou.kurobaexlite.R
import com.github.k1rakishou.kurobaexlite.features.posts.thread.ThreadScreen
import com.github.k1rakishou.kurobaexlite.features.posts.thread.ThreadScreenViewModel
import com.github.k1rakishou.kurobaexlite.helpers.AppConstants
import com.github.k1rakishou.kurobaexlite.helpers.errorMessageOrClassName
import com.github.k1rakishou.kurobaexlite.helpers.image.GrayscaleTransformation
import com.github.k1rakishou.kurobaexlite.helpers.logcatError
import com.github.k1rakishou.kurobaexlite.model.data.local.bookmark.ThreadBookmark
import com.github.k1rakishou.kurobaexlite.model.data.ui.DrawerVisibility
import com.github.k1rakishou.kurobaexlite.model.data.ui.bookmarks.ThreadBookmarkStatsUi
import com.github.k1rakishou.kurobaexlite.model.data.ui.bookmarks.ThreadBookmarkUi
import com.github.k1rakishou.kurobaexlite.model.descriptors.ThreadDescriptor
import com.github.k1rakishou.kurobaexlite.navigation.NavigationRouter
import com.github.k1rakishou.kurobaexlite.themes.ChanTheme
import com.github.k1rakishou.kurobaexlite.ui.elements.snackbar.SnackbarContentItem
import com.github.k1rakishou.kurobaexlite.ui.elements.snackbar.SnackbarId
import com.github.k1rakishou.kurobaexlite.ui.elements.snackbar.SnackbarInfo
import com.github.k1rakishou.kurobaexlite.ui.helpers.KurobaComposeIcon
import com.github.k1rakishou.kurobaexlite.ui.helpers.KurobaComposeText
import com.github.k1rakishou.kurobaexlite.ui.helpers.LazyColumnWithFastScroller
import com.github.k1rakishou.kurobaexlite.ui.helpers.LocalChanTheme
import com.github.k1rakishou.kurobaexlite.ui.helpers.LocalWindowInsets
import com.github.k1rakishou.kurobaexlite.ui.helpers.PullToRefresh
import com.github.k1rakishou.kurobaexlite.ui.helpers.PullToRefreshState
import com.github.k1rakishou.kurobaexlite.ui.helpers.base.ComposeScreen
import com.github.k1rakishou.kurobaexlite.ui.helpers.base.ScreenKey
import com.github.k1rakishou.kurobaexlite.ui.helpers.kurobaClickable
import com.github.k1rakishou.kurobaexlite.ui.helpers.modifier.reorder.ReorderableState
import com.github.k1rakishou.kurobaexlite.ui.helpers.modifier.reorder.detectReorder
import com.github.k1rakishou.kurobaexlite.ui.helpers.modifier.reorder.draggedItem
import com.github.k1rakishou.kurobaexlite.ui.helpers.modifier.reorder.rememberReorderState
import com.github.k1rakishou.kurobaexlite.ui.helpers.modifier.reorder.reorderable
import com.github.k1rakishou.kurobaexlite.ui.helpers.rememberPullToRefreshState
import kotlin.time.Duration.Companion.milliseconds
import kotlinx.coroutines.delay
import kotlinx.coroutines.flow.collectLatest
import org.koin.androidx.viewmodel.ext.android.viewModel

class BookmarksScreen(
  componentActivity: ComponentActivity,
  navigationRouter: NavigationRouter
) : ComposeScreen(componentActivity, navigationRouter) {
  private val bookmarksScreenViewModel: BookmarksScreenViewModel by componentActivity.viewModel()
  private val threadScreenViewModel: ThreadScreenViewModel by componentActivity.viewModel()

  private val circleCropTransformation = CircleCropTransformation()
  private val grayscaleTransformation = GrayscaleTransformation()

  private val deleteBookmarkIconWidth = 40.dp

  override val screenKey: ScreenKey = SCREEN_KEY

  @Composable
  override fun Content() {
    val context = LocalContext.current

    LaunchedEffect(
      key1 = Unit,
      block = {
        snackbarManager.snackbarElementsClickFlow.collectLatest { snackbarClickable ->
          if (snackbarClickable.key !is SnackbarButton) {
            return@collectLatest
          }

          when (snackbarClickable.key as SnackbarButton) {
            SnackbarButton.UndoThreadBookmarkDeletion -> {
              val pair = snackbarClickable.data as? Pair<Int, ThreadBookmark>
                ?: return@collectLatest

              val prevPosition = pair.first
              val threadBookmark = pair.second

              bookmarksScreenViewModel.undoBookmarkDeletion(threadBookmark, prevPosition)
            }
          }
        }
      })

    DisposableEffect(
      key1 = Unit,
      effect = {
        onDispose { bookmarksScreenViewModel.clearMarkedBookmarks() }
      }
    )

    val lazyListState = rememberLazyListState()
    val reorderableState = rememberReorderState(lazyListState = lazyListState)
    val pullToRefreshState = rememberPullToRefreshState()

    LaunchedEffect(
      key1 = Unit,
      block = {
        bookmarksScreenViewModel.backgroundWatcherEventsFlow.collect {
          pullToRefreshState.stopRefreshing()
        }
      }
    )

    BookmarksList(
      pullToRefreshState = pullToRefreshState,
      context = context,
      reorderableState = reorderableState,
    )
  }

  @Composable
  private fun BookmarksList(
    pullToRefreshState: PullToRefreshState,
    context: Context,
    reorderableState: ReorderableState,
  ) {
    val chanTheme = LocalChanTheme.current

    val windowInsets = LocalWindowInsets.current
    val contentPadding = remember(key1 = windowInsets) {
      PaddingValues(top = windowInsets.top, bottom = windowInsets.bottom)
    }
    val pullToRefreshToPadding = remember(key1 = contentPadding) {
      contentPadding.calculateTopPadding()
    }

    val bookmarkList = bookmarksScreenViewModel.bookmarksList
    val canUseFancyAnimations by bookmarksScreenViewModel.canUseFancyAnimations

    PullToRefresh(
      pullToRefreshState = pullToRefreshState,
      topPadding = pullToRefreshToPadding,
      onTriggered = { bookmarksScreenViewModel.forceRefreshBookmarks(context) }
    ) {
      LazyColumnWithFastScroller(
        lazyListContainerModifier = Modifier
          .fillMaxSize()
          .background(chanTheme.backColorCompose),
        lazyListModifier = Modifier
          .fillMaxSize()
          .reorderable(
            state = reorderableState,
            onMove = { from, to -> bookmarksScreenViewModel.onMove(from, to) },
            onDragEnd = { from, to -> bookmarksScreenViewModel.onMoveConfirmed(from, to) }
          ),
        lazyListState = reorderableState.lazyListState,
        contentPadding = contentPadding,
        content = {
          items(
            count = bookmarkList.size,
            key = { index -> bookmarkList[index].threadDescriptor },
            itemContent = { index ->
              val threadBookmarkUi = bookmarkList[index]

              ThreadBookmarkItem(
                canUseFancyAnimations = canUseFancyAnimations,
                threadBookmarkUi = threadBookmarkUi,
                reorderableState = reorderableState,
                onBookmarkClicked = { clickedThreadBookmarkUi ->
                  threadScreenViewModel.loadThread(clickedThreadBookmarkUi.threadDescriptor)
                  globalUiInfoManager.updateCurrentPage(ThreadScreen.SCREEN_KEY)
                  globalUiInfoManager.closeDrawer(withAnimation = true)
                },
                onBookmarkDeleted = { clickedThreadBookmarkUi ->
                  bookmarksScreenViewModel.deleteBookmark(
                    threadDescriptor = clickedThreadBookmarkUi.threadDescriptor,
                    onBookmarkDeleted = { deletedBookmark, oldPosition ->
                      showRevertBookmarkDeletion(
                        context = context,
                        deletedBookmark = deletedBookmark,
                        oldPosition = oldPosition
                      )
                    }
                  )
                },
              )
            }
          )
        }
      )
    }
  }

  private fun showRevertBookmarkDeletion(
    context: Context,
    deletedBookmark: ThreadBookmark,
    oldPosition: Int
  ) {
    val title = deletedBookmark.title ?: deletedBookmark.threadDescriptor.asReadableString()

    snackbarManager.pushSnackbar(
      SnackbarInfo(
        snackbarId = SnackbarId.ThreadBookmarkRemoved,
        aliveUntil = SnackbarInfo.snackbarDuration(AppConstants.deleteBookmarkTimeoutMs.milliseconds),
        content = listOf(
          SnackbarContentItem.Text(
            context.getString(R.string.thread_bookmarks_screen_removed_bookmark_item_text, title)
          ),
          SnackbarContentItem.Spacer(space = 8.dp),
          SnackbarContentItem.Button(
            key = SnackbarButton.UndoThreadBookmarkDeletion,
            text = context.getString(R.string.undo),
            data = Pair(oldPosition, deletedBookmark)
          ),
          SnackbarContentItem.Spacer(space = 8.dp),
        )
      )
    )
  }

  @OptIn(ExperimentalFoundationApi::class)
  @Composable
  private fun LazyItemScope.ThreadBookmarkItem(
    canUseFancyAnimations: Boolean,
    threadBookmarkUi: ThreadBookmarkUi,
    reorderableState: ReorderableState,
    onBookmarkClicked: (ThreadBookmarkUi) -> Unit,
    onBookmarkDeleted: (ThreadBookmarkUi) -> Unit
  ) {
    val chanTheme = LocalChanTheme.current
    val itemHeight = dimensionResource(id = R.dimen.history_or_bookmark_item_height)
    val animationDurationMs = 500

    val isDrawerCurrentlyOpened by listenForDrawerVisibilityEvents()

    val textAnimationSpec = remember(key1 = isDrawerCurrentlyOpened) {
      if (isDrawerCurrentlyOpened && canUseFancyAnimations) {
        tween<Int>(durationMillis = animationDurationMs)
      } else {
        snap<Int>()
      }
    }

    val defaultBgColor = if (bookmarksScreenViewModel.bookmarksToMark.containsKey(threadBookmarkUi.threadDescriptor)) {
      chanTheme.accentColorCompose.copy(alpha = 0.3f)
    } else {
      chanTheme.backColorCompose
    }

    val selectedOnBackColorCompose = remember(key1 = chanTheme.selectedOnBackColorCompose) {
      chanTheme.selectedOnBackColorCompose.copy(alpha = 0.5f)
    }

    var threadBookmarkHash by remember { mutableStateOf(threadBookmarkUi.hashCode()) }
    val bgAnimatable = remember { Animatable(defaultBgColor) }

    LaunchedEffect(
      key1 = threadBookmarkHash,
      key2 = threadBookmarkUi.hashCode(),
      block = {
        if (threadBookmarkUi.hashCode() == threadBookmarkHash) {
          return@LaunchedEffect
        }

        try {
          bgAnimatable.animateTo(selectedOnBackColorCompose, tween(250))
          delay(500)
          bgAnimatable.animateTo(defaultBgColor, tween(250))
        } finally {
          bgAnimatable.snapTo(defaultBgColor)
          threadBookmarkHash = threadBookmarkUi.hashCode()
        }
      }
    )

    val bookmarkBgColor by bgAnimatable.asState()

    Row(
      modifier = Modifier
        .fillMaxWidth()
        .padding(vertical = 2.dp)
        .height(itemHeight)
        .draggedItem(reorderableState.offsetByKey(threadBookmarkUi.threadDescriptor))
        .kurobaClickable(onClick = { onBookmarkClicked(threadBookmarkUi) })
        .drawBehind { drawRect(bookmarkBgColor) }
        .animateItemPlacement(),
      verticalAlignment = Alignment.CenterVertically
    ) {
      KurobaComposeIcon(
        modifier = Modifier
          .size(deleteBookmarkIconWidth)
          .padding(2.dp)
          .kurobaClickable(
            bounded = false,
            onClick = { onBookmarkDeleted(threadBookmarkUi) }
          ),
        drawableId = R.drawable.ic_baseline_close_24
      )

      Spacer(modifier = Modifier.width(4.dp))

      val threadBookmarkStatsUi = threadBookmarkUi.threadBookmarkStatsUi
      val isArchived by threadBookmarkStatsUi.isArchived
      val isDeleted by threadBookmarkStatsUi.isDeleted
      val isDead = isArchived || isDeleted

      if (threadBookmarkUi.thumbnailUrl != null) {
        val thumbnailSize = dimensionResource(id = R.dimen.history_or_bookmark_thumbnail_size)

        BookmarkThumbnail(
          modifier = Modifier
            .size(thumbnailSize)
            .graphicsLayer { alpha = if (isDead) 0.5f else 1f },
          iconUrl = threadBookmarkUi.thumbnailUrl,
          isDead = isDead
        )

        Spacer(modifier = Modifier.width(8.dp))
      }

      Column(modifier = Modifier.weight(1f)) {
        val textColor = if (isDead) {
          chanTheme.textColorHintCompose
        } else {
          chanTheme.textColorPrimaryCompose
        }

        KurobaComposeText(
          modifier = Modifier
            .fillMaxWidth()
            .weight(0.5f),
          text = threadBookmarkUi.title,
          color = textColor,
          fontSize = 15.sp,
          maxLines = 1,
          overflow = TextOverflow.Ellipsis
        )

        ThreadBookmarkAdditionalInfo(
          modifier = Modifier
            .fillMaxWidth()
            .weight(0.5f),
          threadDescriptor = threadBookmarkUi.threadDescriptor,
          threadBookmarkStatsUi = threadBookmarkStatsUi,
          textAnimationSpecProvider = { textAnimationSpec },
        )
      }

      KurobaComposeIcon(
        modifier = Modifier
          .size(32.dp)
          .padding(all = 4.dp)
          .detectReorder(reorderableState),
        drawableId = R.drawable.ic_baseline_reorder_24
      )

      Spacer(modifier = Modifier.width(4.dp))
    }
  }

  @Composable
  private fun listenForDrawerVisibilityEvents(): State<Boolean> {
    val drawerVisibility by globalUiInfoManager.drawerVisibilityFlow
      .collectAsState(initial = DrawerVisibility.Closed)

    return remember { derivedStateOf { drawerVisibility.isOpened } }
  }

  @Composable
  private fun ThreadBookmarkAdditionalInfo(
    modifier: Modifier,
    threadDescriptor: ThreadDescriptor,
    threadBookmarkStatsUi: ThreadBookmarkStatsUi,
    textAnimationSpecProvider: () -> FiniteAnimationSpec<Int>
  ) {
    val context = LocalContext.current
    val chanTheme = LocalChanTheme.current
    val fontSize = 13.sp

    val transition = updateTransition(
      targetState = threadBookmarkStatsUi,
      label = "Bookmark animation"
    )

    val newPostsAnimated by transition.animateInt(
      label = "New posts text animation",
      transitionSpec = { textAnimationSpecProvider() },
      targetValueByState = { state -> state.newPosts.value }
    )
    val newQuotesAnimated by transition.animateInt(
      label = "New quotes text animation",
      transitionSpec = { textAnimationSpecProvider() },
      targetValueByState = { state -> state.newQuotes.value }
    )
    val totalPostsAnimated by transition.animateInt(
      label = "Total posts text textAnimationSpec",
      transitionSpec = { textAnimationSpecProvider() },
      targetValueByState = { state -> state.totalPosts.value }
    )

    val isFirstFetch by threadBookmarkStatsUi.isFirstFetch
    val totalPages by threadBookmarkStatsUi.totalPages
    val currentPage by threadBookmarkStatsUi.currentPage
    val isBumpLimit by threadBookmarkStatsUi.isBumpLimit
    val isImageLimit by threadBookmarkStatsUi.isImageLimit
    val isArchived by threadBookmarkStatsUi.isArchived
    val isDeleted by threadBookmarkStatsUi.isDeleted
    val isError by threadBookmarkStatsUi.isError
    val isDead = isArchived || isDeleted

    val bookmarkAdditionalInfoText = remember(
      newPostsAnimated,
      newQuotesAnimated,
      totalPostsAnimated,
      isFirstFetch,
      totalPages,
      currentPage,
      isBumpLimit,
      isImageLimit,
      isArchived,
      isDeleted,
      isError,
    ) {
      convertBookmarkStateToText(
        context = context,
        chanTheme = chanTheme,
        threadDescriptor = threadDescriptor,
        newPostsAnimated = newPostsAnimated,
        totalPostsAnimated = totalPostsAnimated,
        newQuotesAnimated = newQuotesAnimated,
        totalPages = totalPages,
        currentPage = currentPage,
        isBumpLimit = isBumpLimit,
        isImageLimit = isImageLimit,
        isDeleted = isDeleted,
        isArchived = isArchived,
        isError = isError,
        isDead = isDead,
        isFirstFetch = isFirstFetch,
      )
    }

    val bookmarkInlinedContent = remember(key1 = isDead) {
      val resultMap = mutableMapOf<String, InlineTextContent>()

      BookmarkAnnotatedContent.values().forEach { bookmarkAnnotatedContent ->
        resultMap[bookmarkAnnotatedContent.id] = InlineTextContent(
          placeholder = Placeholder(fontSize, fontSize, PlaceholderVerticalAlign.Center),
          children = { BookmarkAnnotatedContent.Content(bookmarkAnnotatedContent, isDead) }
        )
      }

      return@remember resultMap
    }

    KurobaComposeText(
      modifier = modifier,
      color = Color.Unspecified,
      fontSize = fontSize,
      text = bookmarkAdditionalInfoText,
      inlineContent = bookmarkInlinedContent
    )
  }

  @Composable
  private fun BookmarkThumbnail(
    modifier: Modifier = Modifier,
    iconUrl: String,
    isDead: Boolean
  ) {
    val context = LocalContext.current

    BoxWithConstraints(modifier = modifier) {
      val density = LocalDensity.current
      val desiredSizePx = with(density) { remember { 24.dp.roundToPx() } }

      val iconHeightDp = with(density) {
        remember(key1 = constraints.maxHeight) {
          desiredSizePx.coerceAtMost(constraints.maxHeight).toDp()
        }
      }
      val iconWidthDp = with(density) {
        remember(key1 = constraints.maxWidth) {
          desiredSizePx.coerceAtMost(constraints.maxWidth).toDp()
        }
      }

      val transformations = remember(key1 = isDead) {
        if (isDead) {
          listOf(circleCropTransformation, grayscaleTransformation)
        } else {
          listOf(circleCropTransformation)
        }
      }

      SubcomposeAsyncImage(
        modifier = Modifier.fillMaxSize(),
        model = ImageRequest.Builder(context)
          .data(iconUrl)
          .crossfade(true)
          .transformations(transformations)
          .size(Size.ORIGINAL)
          .build(),
        contentScale = ContentScale.Crop,
        contentDescription = null,
        content = {
          val state = painter.state

          if (state is AsyncImagePainter.State.Error) {
            logcatError(TAG) {
              "BookmarkThumbnail() url=${iconUrl}, error=${state.result.throwable.errorMessageOrClassName()}"
            }

            KurobaComposeIcon(
              modifier = Modifier
                .size(iconWidthDp, iconHeightDp)
                .align(Alignment.Center),
              drawableId = R.drawable.ic_baseline_warning_24
            )

            return@SubcomposeAsyncImage
          }

          SubcomposeAsyncImageContent()
        }
      )
    }
  }

  private fun convertBookmarkStateToText(
    context: Context,
    chanTheme: ChanTheme,
    threadDescriptor: ThreadDescriptor,
    newPostsAnimated: Int,
    totalPostsAnimated: Int,
    newQuotesAnimated: Int,
    totalPages: Int,
    currentPage: Int,
    isBumpLimit: Boolean,
    isImageLimit: Boolean,
    isDeleted: Boolean,
    isArchived: Boolean,
    isError: Boolean,
    isDead: Boolean,
    isFirstFetch: Boolean
  ): AnnotatedString {
    val defaultTextColor = if (isDead) {
      chanTheme.textColorHintCompose
    } else {
      chanTheme.textColorSecondaryCompose
    }

    return buildAnnotatedString {
      pushStyle(SpanStyle(color = defaultTextColor))

      append("/")
      append(threadDescriptor.catalogDescriptor.boardCode)
      append("/")
      append(AppConstants.TEXT_SEPARATOR)

      if (isFirstFetch) {
        append(context.getString(R.string.bookmark_loading_state))
        return@buildAnnotatedString
      }

      append(
        buildAnnotatedString {
          append(
            buildAnnotatedString {
              if (!isDead && newPostsAnimated > 0) {
                pushStyle(SpanStyle(color = chanTheme.bookmarkCounterNormalColorCompose))
              } else {
                pushStyle(SpanStyle(color = defaultTextColor))
              }

              append(newPostsAnimated.toString())
              append("/")
              append(totalPostsAnimated.toString())
            }
          )

          if (newQuotesAnimated > 0) {
            append(" (")
            append(
              buildAnnotatedString {
                if (!isDead && newQuotesAnimated > 0) {
                  pushStyle(SpanStyle(color = chanTheme.bookmarkCounterHasRepliesColorCompose))
                } else {
                  pushStyle(SpanStyle(color = defaultTextColor))
                }

                append(newQuotesAnimated.toString())
              }
            )
            append(")")
          }
        }
      )

      if (totalPages > 0) {
        if (length > 0) {
          append(AppConstants.TEXT_SEPARATOR)
        }

        append("Pg: ")
        append(
          buildAnnotatedString {
            if (!isDead && currentPage >= totalPages) {
              pushStyle(SpanStyle(color = chanTheme.bookmarkCounterNormalColorCompose))
            } else {
              pushStyle(SpanStyle(color = defaultTextColor))
            }

            append(currentPage.toString())
            append("/")
            append(totalPages.toString())
          }
        )
      }

      if (isBumpLimit) {
        if (length > 0) {
          append(AppConstants.TEXT_SEPARATOR)
        }

        append(
          buildAnnotatedString {
            if (!isDead) {
              pushStyle(SpanStyle(color = chanTheme.bookmarkCounterNormalColorCompose))
            }

            append("BL")
          }
        )
      }

      if (isImageLimit) {
        if (length > 0) {
          append(AppConstants.TEXT_SEPARATOR)
        }

        append(
          buildAnnotatedString {
            if (!isDead) {
              pushStyle(SpanStyle(color = chanTheme.bookmarkCounterNormalColorCompose))
            }

            append("IL")
          }
        )
      }

      if (isDeleted) {
        if (length > 0) {
          append(" ")
        }

        appendInlineContent(BookmarkAnnotatedContent.ThreadDeleted.id)
      } else if (isArchived) {
        if (length > 0) {
          append(" ")
        }

        appendInlineContent(BookmarkAnnotatedContent.ThreadArchived.id)
      }

      if (isError) {
        if (length > 0) {
          append(" ")
        }

        appendInlineContent(BookmarkAnnotatedContent.ThreadError.id)
      }
    }
  }

  enum class SnackbarButton {
    UndoThreadBookmarkDeletion
  }

  enum class BookmarkAnnotatedContent(val id: String) {
    ThreadDeleted("id_thread_deleted"),
    ThreadArchived("id_thread_archived"),
    ThreadError("id_thread_error");

    companion object {
      @Composable
      fun Content(bookmarkAnnotatedContent: BookmarkAnnotatedContent, isDead: Boolean) {
        val iconAlpha = remember(key1 = isDead) { if (isDead) 0.5f else 1f }

        val drawableId = remember(key1 = bookmarkAnnotatedContent) {
          when (bookmarkAnnotatedContent) {
            ThreadDeleted -> R.drawable.trash_icon
            ThreadArchived -> R.drawable.archived_icon
            ThreadError -> R.drawable.error_icon
          }
        }

        Image(
          modifier = Modifier
            .fillMaxSize()
            .graphicsLayer { alpha = iconAlpha },
          painter = painterResource(id = drawableId),
          contentDescription = null
        )
      }
    }
  }

  companion object {
    private const val TAG = "BookmarksScreen"
    val SCREEN_KEY = ScreenKey("BookmarksScreen")
  }
}